# -*- coding: utf-8 -*-
"""
---
"""
#%% Functions
def time_interv(t, collectionDuration, kind='hours'):
    """Function for setting time intervals for OD.
    
    Params
    ---------
    t : str, iso, isot, astropy.time.Time
        orbit deterninating time
    
    collectionDuration : int, float
        duration estimation interval
    
    kind: str
        'hours', 'days'
        
    Return
    ----------
    odDate : datetime
        orbit determinating time
    
    startCollectionDate : datetime
        
    t1, t2 : str, isot
        
    t3, t4 : str, isot
    """
    from datetime import datetime, timedelta
    from astropy.time import Time
    
    if isinstance(t, Time):
        pass
    else:
        try: t = Time(t, format='isot', scale='utc')
        except : t = Time(t, format='iso', scale='utc')
    
    odDate = t.datetime 
    
    if kind=='hours':
        startCollectionDate = Time(odDate + timedelta(hours=-collectionDuration), 
                                   format='datetime').datetime
             
        t1 = Time(odDate + timedelta(hours=-collectionDuration), format='datetime').isot
        t2 = Time(odDate + timedelta(hours=+collectionDuration), format='datetime').isot
        
        t3 = Time(odDate + timedelta(hours=-collectionDuration-48), format='datetime').isot
        t4 = Time(odDate + timedelta(hours=+collectionDuration+48), format='datetime').isot
        
    elif kind == 'days':
        startCollectionDate = Time(odDate + timedelta(hours=-collectionDuration), 
                           format='datetime').datetime

        t1 = Time(odDate + timedelta(days=-collectionDuration), format='datetime').isot
        t2 = Time(odDate + timedelta(days=+collectionDuration), format='datetime').isot

        t3 = Time(odDate + timedelta(days=-collectionDuration-2), format='datetime').isot
        t4 = Time(odDate + timedelta(days=+collectionDuration+2), format='datetime').isot
    else: raise ValueError(u' \'{}\' for inerval not supported.'.format(kind))
    return odDate, startCollectionDate, t1, t2, t3, t4

def norm_cov(Q):
    from numpy import zeros, sqrt
    res = zeros(Q.shape)
    for i in range(Q.shape[0]):
        for j in range(Q.shape[1]):
            if i == j:
                res[i,j] = 1
            else:
                res[i,j] = Q[i,j] / (sqrt(Q[i,i]) * sqrt(Q[j,j]))
    return res   

def station(table_stn_crd, epoch, table_stn_ecc=None, system='slr'):
    """Function for translate stations parameters in pandas.DatafFrame, \
        and translate them from python to java (7 column).
    
    Parameters
    ---------
    epoch : float
        date first measurement in mjd format
    
    system : slr'
    
    Return
    ----------
    stationData: pandas.DataFrame
        Latitude, Longitude, Altitude, x, y, z, OrekitGroundStation
    """
    import numpy as np
    import pandas as pd
    from astropy.time import Time
    from org.orekit.utils import IERSConventions
    from org.orekit.frames import FramesFactory
    from org.orekit.time import TimeScalesFactory
    
    UT1 = TimeScalesFactory.getUT1(IERSConventions.IERS_2010, True)
    UT1_eopHistory = UT1.getEOPHistory()
    
    itrf = FramesFactory.getITRF(IERSConventions.IERS_2010, True)
    
    from org.orekit.models.earth import ReferenceEllipsoid
    from org.hipparchus.geometry.euclidean.threed import Vector3D
    from org.orekit.frames import TopocentricFrame
    from org.orekit.estimation.measurements import GroundStation
    from numpy import rad2deg
    from orekit.pyhelpers import datetime_to_absolutedate
        
    if system == 'slr': 
        wgs84ellipsoid = ReferenceEllipsoid.getWgs84(itrf)
        ecef = itrf
        stn_pos  = table_stn_crd[:,2:5].astype(np.float64)
        stn_vel  = table_stn_crd[:,5:8].astype(np.float64)
        
        refEpoch = table_stn_crd[0,13].astype(float)
        stn_ecc = table_stn_ecc[:,7:10].astype(float)
        
    else: 
        raise ValueError(u'Unknown systems: {}'.format(system))
    
    stationData = pd.DataFrame(columns=['Latitude', 'Longitude', 'Altitude',
                                        'x', 'y', 'z', 'OrekitGroundStation'])
    
    for ind, stn in enumerate(table_stn_crd):
        domes = table_stn_crd[ind,0]
        indexTuple = (stn[0])
            
        yearsSinceEpoch = (epoch - refEpoch) / 365.25
        vel = stn_vel[ind,:]
        pos = stn_pos[ind,:]
        
        x = float(pos[0] +                   # Station coordinates
                  vel[0] * yearsSinceEpoch + # Station displacements
                  stn_ecc[ind,0])            # Station eccentricities
        y = float(pos[1] + 
                  vel[1] * yearsSinceEpoch +
                  stn_ecc[ind,1])
        z = float(pos[2] + 
                  vel[2] * yearsSinceEpoch +
                  stn_ecc[ind,2])
        station_xyz_m = Vector3D(x, y, z)
        geodeticPoint = wgs84ellipsoid.transform(station_xyz_m, ecef, datetime_to_absolutedate(Time(epoch, format='mjd').datetime))
        lon_deg = rad2deg(geodeticPoint.getLongitude())
        lat_deg = rad2deg(geodeticPoint.getLatitude())
        alt_m = geodeticPoint.getAltitude()
        topocentricFrame = TopocentricFrame(wgs84ellipsoid, geodeticPoint, str(indexTuple))
        
        from org.orekit.models.earth.displacement import StationDisplacement
        from org.orekit.models.earth.displacement import TidalDisplacement
        from org.orekit.utils import Constants
        from org.orekit.bodies import CelestialBodyFactory
        
        sun = CelestialBodyFactory.getSun()
        moon = CelestialBodyFactory.getMoon()
        
        solid_disp = TidalDisplacement(Constants.EIGEN5C_EARTH_EQUATORIAL_RADIUS,
                                  Constants.JPL_SSD_SUN_EARTH_PLUS_MOON_MASS_RATIO,
                                  Constants.JPL_SSD_EARTH_MOON_MASS_RATIO,
                                  sun, 
                                  moon,
                                  IERSConventions.IERS_2010, 
                                  False)
              
        displacement = StationDisplacement.cast_(solid_disp)
        UT1 = IERSConventions.IERS_2010
        
        groundStation = GroundStation(topocentricFrame, UT1_eopHistory, displacement)
        # groundStation.getClockOffsetDriver().setSelected(True) # station clock offset
        
        stationData.loc[domes] = [lat_deg, lon_deg, alt_m, x, y, z, groundStation]
    return stationData

#%% Initialize Java
# --- Java VM
import orekit
vm = orekit.initVM()

# --- Path for orekit
from java.io import File
from org.orekit.data import DataProvidersManager
orekit_filename = 'orekit-data'
DM = DataProvidersManager.getInstance()
datafile = File(orekit_filename)
if not datafile.exists():
    print('Directory :', datafile.absolutePath, ' not found')
from orekit.pyhelpers import setup_orekit_curdir, absolutedate_to_datetime
setup_orekit_curdir()

#%% Input data
from orekit.pyhelpers import datetime_to_absolutedate
from astropy.time import Time
import numpy as np

# --- Time
t = Time('2018-07-25T00:00:0.000', format='isot', scale='utc') # Time orbit determination
odDate, startCollectionDate, t1_mes, t2_mes, t1, t2 = time_interv(t, 12, kind='hours')

print('========= Time =========')
print(t1_mes, t2_mes)
print(t1, t2)
print()

# --- Satellite paramaters
satellite_mass = 501.6

# --- SLR 
SLR_mjd =  np.load('SLR_mjd.npy')
SLR_mes = np.load('SLR_mes.npy')
SLR_st  = np.load('SLR_st.npy')
SLR_meteo  = np.load('SLR_meteo.npy')

# --- Jason3 PV (ITRF14)
oSrc_ECEF = np.array([[-3370351.103115292,-6932551.077547165,-375286.4364790152, 
                        2261.002393936658, -744.7404452709337,-6560.529489763094]])

# --- ITRF14 -> GCRF
from org.orekit.utils import TimeStampedPVCoordinates, PVCoordinates
from org.hipparchus.geometry.euclidean.threed import Vector3D
from org.orekit.time import TimeScalesFactory
from org.orekit.frames import FramesFactory, ITRFVersion
from org.orekit.utils import IERSConventions

dat1 = datetime_to_absolutedate(odDate)

gcrf = FramesFactory.getGCRF()
itrf = FramesFactory.getITRF(ITRFVersion.ITRF_2014, IERSConventions.IERS_2010, False)

ecef = itrf
eci = gcrf

crd1_ecef = Vector3D(float(oSrc_ECEF[0,0]), float(oSrc_ECEF[0,1]), float(oSrc_ECEF[0,2]))
vel1_ecef = Vector3D(float(oSrc_ECEF[0,3]), float(oSrc_ECEF[0,4]), float(oSrc_ECEF[0,5]))

# Transform to ECI
pv_ecef1 = PVCoordinates(crd1_ecef, vel1_ecef)
pv_eci1 = ecef.getTransformTo(eci, dat1).transformPVCoordinates(pv_ecef1)

crd1 = pv_eci1.getPosition()
vel1 = pv_eci1.getVelocity()

oSrc_GCRF = TimeStampedPVCoordinates(dat1, crd1, vel1)
oSrc_ITRF = TimeStampedPVCoordinates(dat1, crd1_ecef, vel1_ecef)

print("========= ITRF14 =========")
print(oSrc_ITRF)
print()

print("========= GCRF =========")
print(oSrc_GCRF)
print()

# --- Stations coords
stations_eccentricity  = np.load('stations_eccentricity.npy')
stations_coords  = np.load('stations_coords.npy')

stationData_slr = station(table_stn_crd = stations_coords, 
                          epoch = t.mjd,
                          table_stn_ecc = stations_eccentricity,
                          system='slr'
                          )

#%% Measurements
# --- SLR
import pandas as pd
SLR_id = SLR_st
SLR_mjd = Time(SLR_mjd, format='mjd')
SLR_datetime = SLR_mjd.datetime

slrDataFrame = pd.DataFrame(columns=['station-id', 'range', 'weight',
                                     'T', 'P', 'rh'], index = SLR_datetime[:,0])
for i in range(SLR_id.shape[0]):
    ind = np.where(SLR_mes[i,:]!=0)[0] 
    for j in ind:
        P_slr = 1.0
        slrDataFrame.loc[SLR_datetime[j,0], ['station-id', 'range', 
                                         'weight', 'T', 'P', 'rh']] = [SLR_id[i], 
                                                                       float(SLR_mes[i,j]), 
                                                                       float(P_slr), 
                                                                       float(SLR_meteo[j,2]), 
                                                                       float(SLR_meteo[j,0]), 
                                                                       float(SLR_meteo[j,1])/100]
        
# %% Init sat vector state (init orbit)
from org.orekit.models.earth import ReferenceEllipsoid
from org.orekit.propagation import SpacecraftState
from org.orekit.utils import IERSConventions
from org.orekit.orbits import CartesianOrbit

Ellipsoid = ReferenceEllipsoid.getWgs84(ecef)

Cartesian_orbit = CartesianOrbit(oSrc_GCRF, eci, Ellipsoid.getGM())
initialOrbit = Cartesian_orbit

# Attitude
# from org.orekit.attitudes import LofOffset
# from org.orekit.frames import LOFType
# attitudeProvider = LofOffset(Cartesian_orbit.getFrame(), LOFType.LVLH)

initialState = SpacecraftState(initialOrbit, satellite_mass)
print("========= initialState =========")
print(initialState)
print()

#%% Integrator and propagator
# --- Integrator parameters
prop_min_step = 0.001 # s
prop_max_step = 1000.0 # s 
prop_position_error = 10.0 # m 
   
from org.orekit.propagation.conversion import DormandPrince853IntegratorBuilder
integratorBuilder = DormandPrince853IntegratorBuilder(prop_min_step,
                                                      prop_max_step, 
                                                      prop_position_error)

# --- Propagator parameters
estimator_position_scale = 1.0 # m
from org.orekit.propagation.conversion import NumericalPropagatorBuilder
from org.orekit.orbits import PositionAngle
propagatorBuilder = NumericalPropagatorBuilder(initialOrbit, 
                                               integratorBuilder, 
                                               PositionAngle.MEAN,
                                               estimator_position_scale,
                                               # attitudeProvider,
                                               )
propagatorBuilder.setMass(satellite_mass)

#%% Adding perturbation forces to the propagator
# --- Earth gravity field
from org.orekit.forces.gravity import HolmesFeatherstoneAttractionModel
from org.orekit.forces.gravity.potential import GravityFieldFactory
gravityProvider = GravityFieldFactory.getConstantNormalizedProvider(50, 50)
gravityAttractionModel = HolmesFeatherstoneAttractionModel(ecef, gravityProvider)
propagatorBuilder.addForceModel(gravityAttractionModel)

# --- Solid tides
from org.orekit.bodies import CelestialBodyFactory
moon = CelestialBodyFactory.getMoon()
sun = CelestialBodyFactory.getSun()
venus = CelestialBodyFactory.getVenus()
mars = CelestialBodyFactory.getMars()
jupiter = CelestialBodyFactory.getJupiter()

from org.orekit.forces.gravity import SolidTides
solid_tides = SolidTides(Ellipsoid.getBodyFrame(),
                         gravityProvider.getAe(),
                         gravityProvider.getMu(),
                         gravityProvider.getTideSystem(),
                         IERSConventions.IERS_2010,
                         TimeScalesFactory.getUT1(IERSConventions.IERS_2010, True),
                         [sun, moon])
propagatorBuilder.addForceModel(solid_tides)

# --- Ocean Tides
from org.orekit.forces.gravity import OceanTides
order = 34; degree = 34
ocean_tides = OceanTides(Ellipsoid.getBodyFrame(),
                         gravityProvider.getAe(),
                         gravityProvider.getMu(),
                         degree, order, 
                         IERSConventions.IERS_2010,
                         TimeScalesFactory.getUT1(IERSConventions.IERS_2010, True))
propagatorBuilder.addForceModel(ocean_tides)

# --- Moon, Sun, Venus, Mars, Jupiter perturbations
from org.orekit.forces.gravity import ThirdBodyAttraction
moon_3dbodyattraction = ThirdBodyAttraction(moon)
propagatorBuilder.addForceModel(moon_3dbodyattraction)

sun_3dbodyattraction = ThirdBodyAttraction(sun)
propagatorBuilder.addForceModel(sun_3dbodyattraction)

venus_3dbodyattraction = ThirdBodyAttraction(venus)
propagatorBuilder.addForceModel(venus_3dbodyattraction)

mars_3dbodyattraction = ThirdBodyAttraction(mars)
propagatorBuilder.addForceModel(mars_3dbodyattraction)

jupiter_3dbodyattraction = ThirdBodyAttraction(jupiter)
propagatorBuilder.addForceModel(jupiter_3dbodyattraction)

#%% Estimation
# --- Estimator parameters
estimator_convergence_thres = 1.0e-2 
estimator_max_evaluations = 35
estimator_max_iterations = 50 

# --- Setting up the estimator
from org.hipparchus.linear import QRDecomposer
matrixDecomposer = QRDecomposer(1e-11)

from org.hipparchus.optim.nonlinear.vector.leastsquares import GaussNewtonOptimizer 
optimizer = GaussNewtonOptimizer(matrixDecomposer, False)

from org.orekit.estimation.leastsquares import BatchLSEstimator
estimator = BatchLSEstimator(optimizer, propagatorBuilder)

estimator.setParametersConvergenceThreshold(estimator_convergence_thres)
estimator.setMaxIterations(estimator_max_iterations)
estimator.setMaxEvaluations(estimator_max_evaluations)

from org.orekit.estimation.measurements import Range, ObservableSatellite
observableSatellite = ObservableSatellite(0)

 
# observableSatellite.getClockOffsetDriver().setMinValue(0.00000000001)
# observableSatellite.getClockOffsetDriver().setMaxValue(0.1)
# observableSatellite.getClockOffsetDriver().setReferenceValue(0.0000001)
# observableSatellite.getClockOffsetDriver().setValue(0.0000001)
observableSatellite.getClockOffsetDriver().setSelected(True)

# --- SLR measurements
from org.orekit.estimation.measurements.modifiers import RangeTroposphericDelayModifier

from org.orekit.estimation.measurements.modifiers import OutlierFilter, ShapiroRangeModifier
rangeOutliersManager = OutlierFilter(2, 1000.)

from org.orekit.models.earth.troposphere import DiscreteTroposphericModel, MendesPavlisModel
range_sigma = 1.0 
for receiveTime, slrData in slrDataFrame.iterrows():    
    orekitRange = Range(stationData_slr.loc[slrData['station-id'], 'OrekitGroundStation'], 
                        True, 
                        datetime_to_absolutedate(receiveTime),
                        slrData['range'],
                        range_sigma,
                        slrData['weight'],
                        observableSatellite
                        ) 

    # Troposphere
    tropo_model_slr = MendesPavlisModel(slrData['T'],                 
                                        slrData['P'],                 
                                        slrData['rh'],                
                                        float(np.deg2rad(stationData_slr.loc[slrData['station-id'],'Latitude'])),
                                        float(0.532),                  
                                        )
    diskr_model_slr = DiscreteTroposphericModel.cast_( tropo_model_slr )
    orekitRange.addModifier(RangeTroposphericDelayModifier(diskr_model_slr))
    
    # Shapiro
    orekitRange.addModifier(ShapiroRangeModifier(initialOrbit.getMu())) 
    
    # OutliersManager
    orekitRange.addModifier(rangeOutliersManager)
    
    estimator.addMeasurement(orekitRange)


#==========================================================================      
test2 = estimator.getMeasurementsParametersDrivers(False).drivers
test3 = estimator.getPropagatorParametersDrivers(True).drivers
test4 = estimator.getOrbitalParametersDrivers(True).drivers
print(test2, test3, test4)
#============================================================================

#%% Performing the OD
import datetime
print('=========== estimating ===========')
print(datetime.datetime.now(),'Start estimating...')
estimatedPropagatorArray = estimator.estimate()
print(datetime.datetime.now(), 'Estimating done')
print()

print('sat clock offset: {}'.format(observableSatellite.getClockOffsetDriver().getValue()))
print()

#%% Propagating the estimated orbit
dt = 60.0

date_start = datetime_to_absolutedate(startCollectionDate)
date_start = date_start.shiftedBy(-86400.0)
date_end = datetime_to_absolutedate(odDate)
date_end = date_end.shiftedBy(86400.0)

# First propagating in ephemeris mode
estimatedPropagator = estimatedPropagatorArray[0]
estimatedInitialState = estimatedPropagator.getInitialState()
actualOdDate = estimatedInitialState.getDate()
estimatedPropagator.resetInitialState(estimatedInitialState)
estimatedPropagator.setEphemerisMode()

# Propagating from 1 day before data collection
estimatedPropagator.propagate(date_start, datetime_to_absolutedate(odDate).shiftedBy(7 * 86400.0))
bounded_propagator = estimatedPropagator.getGeneratedEphemeris()

#%% Covariance analysis
# Creating the LVLH (local vertical local horizontal) frame, 
# computing the covariance matrix in both TOD and LVLH frames

# Creating the LVLH frame 
# It must be associated to the bounded propagator, not the original numerical propagator
from org.orekit.frames import LocalOrbitalFrame
from org.orekit.frames import LOFType
lvlh = LocalOrbitalFrame(eci, LOFType.LVLH, bounded_propagator, 'LVLH')

# Getting covariance matrix in ECI frame
covMat_eci_java = estimator.getPhysicalCovariances(1.0e-10)

# Converting matrix to LVLH frame
"""
Getting an inertial frame aligned with the LVLH frame at this instant
The LVLH is normally not inertial, but this should not affect results too much
Reference: David Vallado, Covariance Transformations for Satellite Flight Dynamics Operations, 2003
"""
eci2lvlh_frozen = eci.getTransformTo(lvlh, actualOdDate).freeze() 
eci2ecef_frozen = eci.getTransformTo(ecef, actualOdDate).freeze() 

from org.orekit.utils import CartesianDerivativesFilter
from orekit.pyhelpers import JArray_double2D
from org.hipparchus.linear import Array2DRowRealMatrix

n=6 #!!!
# lvlh
# Computing Jacobian
jacobianDoubleArray = JArray_double2D(n, n)
eci2lvlh_frozen.getJacobian(CartesianDerivativesFilter.USE_PV, jacobianDoubleArray)
jacobian = Array2DRowRealMatrix(jacobianDoubleArray)
# Applying Jacobian to convert matrix to lvlh
covMat_lvlh_java = jacobian.multiply(covMat_eci_java.multiply(jacobian.transpose()))

# ecef 
# Computing Jacobian
jacobianDoubleArray = JArray_double2D(n, n)
eci2ecef_frozen.getJacobian(CartesianDerivativesFilter.USE_PV, jacobianDoubleArray)
jacobian = Array2DRowRealMatrix(jacobianDoubleArray)
# Applying Jacobian to convert matrix to ecef
covMat_ecef_java = jacobian.multiply(covMat_eci_java.multiply(jacobian.transpose()))

# Converting the Java matrices to numpy
import numpy as np
covarianceMat_eci  = np.matrix([covMat_eci_java.getRow(iRow) 
                              for iRow in range(0, covMat_eci_java.getRowDimension())])
covarianceMat_lvlh = np.matrix([covMat_lvlh_java.getRow(iRow) 
                              for iRow in range(0, covMat_lvlh_java.getRowDimension())])
covarianceMat_ecef = np.matrix([covMat_ecef_java.getRow(iRow) 
                              for iRow in range(0, covMat_ecef_java.getRowDimension())])

# Computing the position and velocity standard deviation
print("=========== СКО_LVLH ===========")
pos_std_crossTrack = np.sqrt(covarianceMat_lvlh[0,0])
pos_std_alongTrack = np.sqrt(covarianceMat_lvlh[1,1])
pos_std_outOfPlane = np.sqrt(covarianceMat_lvlh[2,2])
print(f'Position std: \n cross-track {pos_std_crossTrack:.3e} m, \n along-track {pos_std_alongTrack:.3e} m, \n out-of-plane {pos_std_outOfPlane:.3e} m')
vel_std_crossTrack = np.sqrt(covarianceMat_lvlh[3,3])
vel_std_alongTrack = np.sqrt(covarianceMat_lvlh[4,4])
vel_std_outOfPlane = np.sqrt(covarianceMat_lvlh[5,5])
print(f'Velocity std: \n cross-track {vel_std_crossTrack:.3e} m/s, \n along-track {vel_std_alongTrack:.3e} m/s, \n out-of-plane {vel_std_outOfPlane:.3e} m/s')
print()

print("=========== СКО_ECI ===========")
pos_std_crossTrack = np.sqrt(covarianceMat_eci[0,0])
pos_std_alongTrack = np.sqrt(covarianceMat_eci[1,1])
pos_std_outOfPlane = np.sqrt(covarianceMat_eci[2,2])
print(f'Position std: \n X {pos_std_crossTrack:.3e} m, \n Y {pos_std_alongTrack:.3e} m, \n Z {pos_std_outOfPlane:.3e} m')
vel_std_crossTrack = np.sqrt(covarianceMat_eci[3,3])
vel_std_alongTrack = np.sqrt(covarianceMat_eci[4,4])
vel_std_outOfPlane = np.sqrt(covarianceMat_eci[5,5])
print(f'Velocity std: \n Vx {vel_std_crossTrack:.3e} m/s, \n Vy {vel_std_alongTrack:.3e} m/s, \n Vz {vel_std_outOfPlane:.3e} m/s')
print()

print("=========== СКО_ECEF ===========")
pos_std_crossTrack = np.sqrt(covarianceMat_ecef[0,0])
pos_std_alongTrack = np.sqrt(covarianceMat_ecef[1,1])
pos_std_outOfPlane = np.sqrt(covarianceMat_ecef[2,2])
print(f'Position std: \n X {pos_std_crossTrack:.3e} m, \n Y {pos_std_alongTrack:.3e} m, \n Z {pos_std_outOfPlane:.3e} m')
vel_std_crossTrack = np.sqrt(covarianceMat_ecef[3,3])
vel_std_alongTrack = np.sqrt(covarianceMat_ecef[4,4])
vel_std_outOfPlane = np.sqrt(covarianceMat_ecef[5,5])
print(f'Velocity std: \n Vx {vel_std_crossTrack:.3e} m/s, \n Vy {vel_std_alongTrack:.3e} m/s, \n Vz {vel_std_outOfPlane:.3e} m/s')
print()

#%% Analyzing residuals
propagatorParameters   = estimator.getPropagatorParametersDrivers(True)
measurementsParameters = estimator.getMeasurementsParametersDrivers(True)

lastEstimations = estimator.getLastEstimations()
valueSet = lastEstimations.values()
estimatedMeasurements = valueSet.toArray()
keySet = lastEstimations.keySet()
realMeasurements = keySet.toArray()

from org.orekit.estimation.measurements import EstimatedMeasurement

import pandas as pd
residuals = pd.DataFrame(columns=['residuals'])

range_est = pd.DataFrame(columns=['range_est'])
range_real = pd.DataFrame(columns=['range_real'])
range_residuals = pd.DataFrame(columns=['range'])

for estMeas, realMeas in zip(estimatedMeasurements, realMeasurements):
    estMeas = EstimatedMeasurement.cast_(estMeas)
    estimatedValue = estMeas.getEstimatedValue()
    pyDateTime = absolutedate_to_datetime(estMeas.date)
    
    if Range.instance_(realMeas):
        observedValue = Range.cast_(realMeas).getObservedValue()
        range_residuals.loc[pyDateTime] = np.array(observedValue) - np.array(estimatedValue)
        residuals.loc[pyDateTime] = np.array(observedValue) - np.array(estimatedValue)
        
        range_est.loc[pyDateTime] = np.array(estimatedValue)
        range_real.loc[pyDateTime] = np.array(observedValue)
        
# --- Plotting range residuals
import matplotlib.pyplot as plt
est_res = estimator.getOptimum().getResiduals().toArray()
plt.figure(figsize=(16,8))
plt.title('Init inaccuracy')
plt.plot(est_res, 'r.')
plt.xlabel('Equation', fontsize=14)
plt.ylabel('Inaccuracy, м', fontsize=14)
plt.grid()

plt.figure(figsize=(16,8))
plt.title('Range residuals')
plt.plot(np.array(range_residuals['range']), 'r.')
plt.xlabel('Equation', fontsize=14)
plt.ylabel('Residual, м', fontsize=14)
plt.grid()

#%% Statistics
N_eq_slr = range_real.shape[0]
STDR_slr = np.array(range_residuals['range']).std() 
SSR_slr  = np.sum(np.array(range_residuals['range'])**2) 

N_it = estimator.getIterationsCount() 

print('Number of measurements SLR: {}'.format( N_eq_slr ))
print('Number of iters: {}'.format( N_it ))
print('Sum squares of residuals SLR: {}'.format( SSR_slr ))
print('error of unit weight SLR: {}'.format( SSR_slr/(N_eq_slr-6) ))
print('STD residuals SLR: {}'.format(STDR_slr))
print()

# --- Norm cov matrix
norm_cov_LVLH = norm_cov(covarianceMat_lvlh)
norm_cov_ECEF = norm_cov(covarianceMat_ecef)
with np.printoptions(precision=3):
    print("==== Norm cov matrix in ITRF14 ====")
    print(norm_cov_ECEF)
    print()